from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import List, Optional


@dataclass_json
@dataclass
class InstallConfig:
    run: List[str]
    cwd: Optional[str] = None



import logging
from typing import Callable

from lume.src.domain.services.interface_logger import (
    ILogger,
    INFO,
    ERROR,
    WARN,
    WARNING,
    DEBUG,
    CRITICAL,
    FATAL,
)


class LoggingBasedLogger(ILogger):
    def __init__(
        self,
        logger_name: str,
        logging_level=logging.INFO,
        format: str = "%(name)s - %(levelname)s - %(message)s",
        config_func: Callable = None,
    ):

        logging.basicConfig(format=format, level=logging_level)
        if config_func:
            config_func()
        self.logger = logging.getLogger(logger_name)

    def log(self, logging_level, message):
        if logging_level == INFO:
            self.logger.info(message)
        elif logging_level == DEBUG:
            self.logger.debug(message)
        elif logging_level == CRITICAL:
            self.logger.critical(message)
        elif logging_level == FATAL:
            self.logger.fatal(message)
        elif logging_level == ERROR:
            self.logger.error(message)
        elif logging_level == WARN:
            self.logger.warn(message)
        elif logging_level == WARNING:
            self.logger.warning(message)



import os
from typing import List

from meiga import Result, Error, Success, Failure
from meiga.decorators import meiga

from lume.config import Config
from lume.src.domain.services.interface_executor_service import IExecutorService
from lume.src.domain.services.interface_logger import ILogger, INFO, WARNING, ERROR
from lume.src.domain.services.interface_setup_service import ISetupService


class EmptyConfigError(Error):
    pass


class ActionNotFoundError(Error):
    pass


class CwdIsNotADirectoryError(Error):
    pass


def on_empty_config(self, action):
    self.logger.log(WARNING, f"Empty config for {action}")


def on_error_with_cwd(self, action):
    self.logger.log(
        WARNING, f"Given execution directory (cwd) for {action} action is not valid"
    )


class LumeUseCase:
    def __init__(
        self,
        config: Config,
        executor_service: IExecutorService,
        setup_service: ISetupService,
        logger: ILogger,
    ):
        self.config = config
        self.executor_service = executor_service
        self.setup_service = setup_service
        self.logger = logger

    @meiga
    def execute(self, actions: List[str]):

        for action in actions:
            self.logger.log(INFO, f"Action: {action}")

            if action == "setup":
                result = self.setup_service.execute()
                if result.is_failure:
                    self.logger.log(ERROR, f"Setup: {result.value}")
                result.unwrap_or_throw()
            else:
                commands = (
                    self.get_commands(action)
                    .handle(on_failure=on_empty_config, failure_args=(self, action))
                    .unwrap_or([])
                )

                result = self.get_cwd(action).handle(
                    on_failure=on_error_with_cwd, failure_args=(self, action)
                )

                if result.is_failure:
                    continue

                cwd = result.unwrap()
                for command in commands:
                    message = (
                        f"{action} >> {command}"
                        if not cwd
                        else f"{action} [cwd={cwd}] >> {command}"
                    )
                    self.logger.log(INFO, message)
                    self.executor_service.execute(command, cwd).unwrap_or_throw()

    def get_commands(self, action) -> Result[List[str], Error]:
        if action == "install":
            if not self.config.install:
                return Failure(EmptyConfigError())
            commands = self.config.install.run
        else:
            step = self.config.steps.get(action)
            if not step:
                return Failure(EmptyConfigError())
            commands = step.run

        return Success(commands)

    def get_cwd(self, action) -> Result[str, Error]:
        if action == "install":
            if not self.config.install:
                return Success(None)
            cwd = self.config.install.cwd
            if cwd and not os.path.isdir(cwd):
                return Failure(CwdIsNotADirectoryError())
            return Success(cwd)
        else:
            if not self.config.steps:
                return Success(None)
            step = self.config.steps.get(action)
            if step:
                cwd = step.cwd
                if cwd and not os.path.isdir(cwd):
                    return Failure(CwdIsNotADirectoryError())
                return Success(cwd)
            else:
                return Failure(ActionNotFoundError())



import emoji
from lume.src.domain.services.interface_logger import ILogger, INFO, WARNING, ERROR

LOGGING_LEVEL = {
    INFO: emoji.emojize("👩‍💻 "),
    ERROR: emoji.emojize("❌ "),
    WARNING: emoji.emojize("🧐 "),
}


class EmojisLogger(ILogger):
    def log(self, logging_level, message):
        print(f"{LOGGING_LEVEL[logging_level]}>> {message.rstrip()}")



from subprocess import Popen, PIPE
from typing import Dict

from meiga import Result, Error, isSuccess, isFailure

from lume.src.domain.services.interface_executor_service import IExecutorService

# TODO pythonize this function
# It's possible to add a custom "PIPE" handler to the logger and pass that handler to
# Popen's stdout and stderr
from lume.src.domain.services.interface_logger import ILogger, WARNING, ERROR, INFO


def get_and_log_process_std(process, logger):
    """ Capture and log the stdout and stderr of a running process
    This allows process that take a while to finish (i.e. pytest) to avoid waiting for
    process.communicate() and log intermediate outputs.
    """
    output = err = ""
    while True:
        current_output = process.stdout.readline().decode()
        current_err = None
        if current_output:
            logger.debug(current_output)
            output += current_output
        else:
            current_err = process.stderr.readline().decode()
            if current_err:
                logger.warning(current_err)
                err += current_err
        if not current_output and not current_err:
            break
    return output, err


class PopenExecutorService(IExecutorService):
    def __init__(
        self, logger: ILogger, use_communicate=True, raise_runtime_error=False
    ):
        self.logger = logger
        self.use_communicate = use_communicate
        self.raise_runtime_error = raise_runtime_error

    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def execute(self, command: str, cwd: str) -> Result[bool, Error]:

        if not cwd:
            cwd = "."

        process = Popen(command, stdout=PIPE, stderr=PIPE, cwd=cwd, shell=True)

        if self.use_communicate:
            output, err = process.communicate()
        else:
            output, err = get_and_log_process_std(process, self.logger)

        if isinstance(output, bytes):
            output = output.decode("utf-8")
        if isinstance(err, bytes):
            err = err.decode("utf-8")
        return_code = process.poll()

        output = output.rstrip()

        if output != "":
            self.logger.log(INFO, f"{output}")

        if return_code == 0:
            if err:
                self.logger.log(WARNING, err)
        else:
            # something weird happened
            if err:
                self.logger.log(ERROR, err)
            # but may have not been written to stderr
            # i.e. flake8 fails with return code 1 but writes to stout
            else:
                self.logger.log(ERROR, output)

            if self.raise_runtime_error:
                raise RuntimeError(
                    "Command '{}' has failed with return_code '{}'".format(
                        command, return_code
                    )
                )
            return isFailure

        return isSuccess



from lume.config import Config
from lume.src.application.use_cases.lume_use_case import LumeUseCase
from lume.src.infrastructure.services.executor.popen_executor_service import (
    PopenExecutorService,
)
from lume.src.infrastructure.services.logger.emojis_logger import EmojisLogger
from lume.src.infrastructure.services.setup.setup_service import SetupService


class UseCaseBuilder:
    @staticmethod
    def lume(config: Config):

        default_logger = EmojisLogger()
        default_executor_service = PopenExecutorService(logger=default_logger)
        default_setup_service = SetupService(
            setup_config=config.setup, logger=default_logger
        )

        return LumeUseCase(
            config=config,
            executor_service=default_executor_service,
            setup_service=default_setup_service,
            logger=default_logger,
        )



from dataclasses import dataclass
from typing import Dict, List, Optional

from dataclasses_json import dataclass_json

from lume.config.install_config import InstallConfig
from lume.config.setup_config import SetupConfig
from lume.config.step_config import StepConfig


@dataclass_json
@dataclass
class Config:
    name: Optional[str] = None
    install: Optional[InstallConfig] = None
    setup: Optional[SetupConfig] = None
    steps: Optional[Dict[str, StepConfig]] = None

    def get_steps(self) -> List[str]:
        return list(self.steps.keys())

    def get_commands(self) -> List[str]:
        commands = []
        if self.install:
            commands.append("install")
        if self.setup:
            commands.append("setup")
        commands += self.get_steps()
        return commands



from abc import abstractmethod, ABCMeta

from meiga import Result

from lume.config import DependencyConfig
from lume.src.domain.services.interface_logger import ILogger


class SetupItem(object):
    __metaclass__ = ABCMeta
    base_path = None

    def __init__(self, base_path: str):
        self.base_path = base_path

    @abstractmethod
    def run(
        self, name: str, dependency_config: DependencyConfig, logger: ILogger
    ) -> Result:
        raise NotImplementedError
